#ifndef __UNIX_BUFFER__
#define __UNIX_BUFFER__


extern "C"
{
  #include <sys/socket.h>
  #include <netinet/in.h>
  #include <netdb.h>
  #include <unistd.h>
  #include <errno.h>
};

#include <string>
#include <iostream>

using std::cerr;
using std::endl;
using std::flush;
using std::iostream;
using std::streambuf;

class unix_buffer : public streambuf {
protected:
  static const int buffersize = 4096;
  static const int pushbacksize = 10;
  int fd;
  bool is_open;
  char obuffer[buffersize];
  char ibuffer[buffersize+pushbacksize];

  const int empty_buffer(void)
   {
     if (fd<0) // invalid file descriptor?
      {
        cerr << "unix_buffer::empty buffer(): invalid descriptor! fd=" << fd << endl;
        return EOF;
      }
     int count = pptr()-pbase(); if (!count) return 0;
     int written=send(fd, obuffer, count, MSG_NOSIGNAL);
     if (written!=count)
      {
        cerr << "unix_buffer::empty_buffer()::"; perror(NULL);
        if (written<=0) return EOF;
      }
     pbump(-written);
     return written;
   }

  virtual int overflow (int c) 
   {
     if (c != EOF) { *pptr() = c; pbump(1); }
     if (empty_buffer() == EOF) return EOF;
     return c;
   }

  virtual int sync ()
   {
     return (empty_buffer() == EOF) ? -1 : 0;
   }

  virtual int underflow ()
   {
     if (gptr() < egptr()) return *gptr();
     int pbcount = gptr() - eback();
     if (pbcount > pushbacksize) pbcount = pushbacksize;
     if (pbcount > 0) memmove (ibuffer+(pushbacksize-pbcount), gptr()-pbcount, pbcount);

     int count;
     do {
       count = recv(fd, ibuffer+pushbacksize, buffersize, MSG_NOSIGNAL);
       if (count <= 0) // any unexpected behaviour?
        {
         if (count==0) return EOF;
         if (count==-1 && errno == EINTR) continue; // this should fix some problems...
 	 perror(NULL); return EOF;
        }
     } while (count<=0);
     setg(ibuffer+(pushbacksize-pbcount), ibuffer+pushbacksize,
	  ibuffer+pushbacksize+count);
     return *gptr();
   }
  
public:

  inline void set_timeout (const int secs, const int u_secs = 0)
   {
     timeval tv;
     tv.tv_sec = secs; // secs timeout
     tv.tv_usec = u_secs; // micro-secs timeout
     if (setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv))==-1)
       { cerr << "unix_buffer::set_timeout() "; perror(NULL); }
     if (setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv))==-1)
       { cerr << "unix_buffer::set_timeout() "; perror(NULL); }
   }

  void open(const int _fd)
  {
    if (fd>0)
     {
       cerr << "unix_buffer::open(const int _fd): descriptor is already in use!" << endl;
       exit(1);
     }
    if (_fd<0)
     {
       cerr << "unix_buffer::open(const int _fd): descriptor is invalid!" << endl;
       exit(1);
     }
    fd = _fd;
    is_open = true;
    setp(obuffer, obuffer+(buffersize-1));
    setg(ibuffer+pushbacksize, ibuffer+pushbacksize, ibuffer+pushbacksize);
  }

  unix_buffer(void) // Default constructor
   : fd(0), // 0 denotes no descriptor defined yet
     is_open(true) // as no descriptor is defined, no descriptor need to be closed later
  {
    set_timeout(75); // Standard: 75 Sekunden
    setp(obuffer, obuffer+(buffersize-1));
    setg(ibuffer+pushbacksize, ibuffer+pushbacksize, ibuffer+pushbacksize);
  }

  unix_buffer(const int _fd, const int secs_timeout = 75)
   : fd(_fd), is_open(true)
  {
    set_timeout(secs_timeout);
    setp(obuffer, obuffer+(buffersize-1));
    setg(ibuffer+pushbacksize, ibuffer+pushbacksize, ibuffer+pushbacksize);
  }


 unix_buffer(const char * const host, int port)
  : fd(0), is_open(true)
  {
    struct sockaddr_in server;
    struct hostent *hp;
    int ret;
    int err_number;
    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0)
     {
      cerr << "Can't open socket" << endl;
      exit(1);
     }
    server.sin_family = AF_INET;
    hp = gethostbyname(host);
    if (hp == NULL)
     {
       cerr << "Unknown host: " << host << endl;
       exit(1);
     }
    memcpy((char*)&server.sin_addr, (char*)hp->h_addr, hp->h_length);

    server.sin_port = htons(port);
    {
      int retries_on_EAGAIN = 3;
      while ((ret = connect(fd, (struct sockaddr *)&server, sizeof(server))) < 0)
       {
	switch (err_number=errno)
         {
	   case ETIMEDOUT:
            cerr << "." << flush; sleep(10); break;
	   case ENETUNREACH:
            cerr << "%" << flush; sleep(10); break;
           case EAGAIN:
            if (retries_on_EAGAIN--)
             {
               cerr << "Can't connect socket (" << strerror(err_number) << ")" << endl;
               if (err_number==EAGAIN)
                {
                  cerr << "Trying it again in 10 secs..." << endl;
                  sleep(10);
                }
	       break;
             }
	   default:
	    cerr << "Can't connect socket (" << strerror(err_number) << ")" << endl;
            exit(1);
	 }
       }
    }
    is_open = false;
    setp(obuffer, obuffer+(buffersize-1));
    setg(ibuffer+pushbacksize, ibuffer+pushbacksize, ibuffer+pushbacksize);
  }

  virtual ~unix_buffer()
   {
     if (fd<0) return;
     sync();
     if (!is_open)
      {
        shutdown(fd, SHUT_RDWR);
        close(fd);
      }
   }
};

class unix_io_client : public iostream {
protected:
  unix_buffer buf;
public:
  unix_io_client(const char * const host, const int port) : iostream(&buf), buf(host, port) { }
};

int open_internet_port(const int port, const int qsize = 5)
{
  int sd;
  int val = 1;
  struct sockaddr_in server;
  struct linger lingval;

  if (port <= 0) return -1;

  sd = socket(AF_INET, SOCK_STREAM, 0);
  if (sd < 0) return sd;
  if (setsockopt(sd, SOL_SOCKET, SO_REUSEADDR, (char *)&val, sizeof(val)) <0)
    return -1;

  // nderung Thorsten: Linger-Option aktivieren
  lingval.l_onoff=1; // beim Schlieen des Sockets ggf. warten
  lingval.l_linger=10*100; // Timeout = 10 Sekunden
  if (setsockopt(sd, SOL_SOCKET, SO_LINGER, (char *)&lingval, sizeof(lingval)) <0)
    return -1;

  server.sin_family = AF_INET;
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_port = htons(port);
  if (bind(sd, (struct sockaddr *)&server, sizeof(server)) < 0) return -1;
  listen (sd, qsize);
  return sd;
}

#endif
